/*                                                                              */
/* Copyright (C) 2001 International Business Machines                           */
/* All rights reserved.                                                         */
/*                                                                              */
/* This file is part of the GPFS user library.                                  */
/*                                                                              */
/* Redistribution and use in source and binary forms, with or without           */
/* modification, are permitted provided that the following conditions           */
/* are met:                                                                     */
/*                                                                              */
/*  1. Redistributions of source code must retain the above copyright notice,   */
/*     this list of conditions and the following disclaimer.                    */
/*  2. Redistributions in binary form must reproduce the above copyright        */
/*     notice, this list of conditions and the following disclaimer in the      */
/*     documentation and/or other materials provided with the distribution.     */
/*  3. The name of the author may not be used to endorse or promote products    */
/*     derived from this software without specific prior written                */
/*     permission.                                                              */
/*                                                                              */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR         */
/* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES    */
/* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.      */
/* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;  */
/* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,     */
/* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR      */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF       */
/* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                   */
/*                                                                              */
/* @(#)83       1.10.1.2  src/avs/fs/mmfs/ts/util/gpfs_nfs.h, mmfs, avs_rhin, rhins009a 10/10/11 11:48:44 */
/*
 *  Library calls for GPFS interfaces
 */
#ifndef H_GPFS_NFS
#define H_GPFS_NFS

#ifdef __cplusplus
extern "C" {
#endif

#ifdef WIN32
struct flock
{};
#endif

/* GANESHA common information */

#define OPENHANDLE_NAME_TO_HANDLE 101
#define OPENHANDLE_OPEN_BY_HANDLE 102
#define OPENHANDLE_LINK_BY_FD     103
#define OPENHANDLE_READLINK_BY_FD 104
#define OPENHANDLE_STAT_BY_HANDLE 105
#define OPENHANDLE_LAYOUT_TYPE    106
#define OPENHANDLE_GET_DEVICEINFO 107
#define OPENHANDLE_GET_DEVICEITER 108
#define OPENHANDLE_LAYOUT_GET     109
#define OPENHANDLE_LAYOUT_RETURN  110
#define OPENHANDLE_INODE_UPDATE   111
#define OPENHANDLE_GET_XSTAT      112
#define OPENHANDLE_SET_XSTAT      113
#define OPENHANDLE_CHECK_ACCESS   114
#define OPENHANDLE_OPEN_SHARE_BY_HANDLE 115
#define OPENHANDLE_GET_LOCK       116
#define OPENHANDLE_SET_LOCK       117
#define OPENHANDLE_THREAD_UPDATE  118

int gpfs_ganesha(int op, void *oarg);

#define OPENHANDLE_HANDLE_LEN 40
#define OPENHANDLE_KEY_LEN 28
struct gpfs_file_handle
{
  int handle_size;
  int handle_type;
  int handle_key_size;
  /* file identifier */
  unsigned char f_handle[OPENHANDLE_HANDLE_LEN];
};

struct name_handle_arg
{
  int dfd;
  int flag;
  char *name;
  struct gpfs_file_handle *handle;
};

struct open_arg
{
  int mountdirfd;
  int flags;
  int openfd;
  struct gpfs_file_handle *handle;
};

struct glock
{
  int cmd;
  int lfd;
  void *lock_owner;
  struct flock flock;
};

struct set_get_lock_arg
{
  int mountdirfd;
  struct glock *lock;
};

struct open_share_arg
{
  int mountdirfd;
  int flags;
  int openfd;
  struct gpfs_file_handle *handle;
  int share_access;
  int share_deny;
};

struct link_arg
{
  int file_fd;
  int dir_fd;
  char *name;
};

struct readlink_arg
{
  int fd;
  char *buffer;
  int size;
};

struct nfsd4_pnfs_deviceid {
	unsigned long	sbid;		/* per-superblock unique ID */
	unsigned long	devid;		/* filesystem-wide unique device ID */
};

struct exp_xdr_stream {
	int *p;
	int *end;
};

struct deviceinfo_arg
{
  int mountdirfd;
  int type;
  struct nfsd4_pnfs_deviceid devid;
  struct exp_xdr_stream xdr;
};

struct stat_arg
{
    int mountdirfd;
    struct gpfs_file_handle *handle;
#if BITS_PER_LONG != 64
    struct stat64 *buf;
#else
    struct stat *buf;
#endif
};

struct callback_arg
{
    int mountdirfd;
    int *reason;
    struct gpfs_file_handle *handle;
    struct glock *fl;
#if BITS_PER_LONG != 64
    struct stat64 *buf;
#else
    struct stat *buf;
#endif
};
/* reason list */
#define INODE_INVALIDATE 1
#define INODE_UPDATE     2
#define INODE_LOCK_GRANTED 3
#define INODE_LOCK_AGAIN   4
#define THREAD_STOP        5
#define THREAD_PAUSE       6

/* define flags for attr_valid */
#define XATTR_STAT      (1 << 0)
#define XATTR_ACL       (1 << 1)

/* define flags for attr_chaged */
#define XATTR_MODE      (1 << 0) //  01
#define XATTR_UID       (1 << 1) //  02
#define XATTR_GID       (1 << 2) //  04
#define XATTR_SIZE      (1 << 3) //  08
#define XATTR_ATIME     (1 << 4) //  10
#define XATTR_MTIME     (1 << 5) //  20
#define XATTR_CTIME     (1 << 6) //  40
#define XATTR_ATIME_SET (1 << 7) //  80
#define XATTR_MTIME_SET (1 << 8) // 100

struct xstat_arg
{
    int attr_valid;
    int mountdirfd;
    struct gpfs_file_handle *handle;
    struct gpfs_acl *acl;
    int attr_changed;
#if BITS_PER_LONG != 64
    struct stat64 *buf;
#else
    struct stat *buf;
#endif
};

struct xstat_access_arg
{
    int mountdirfd;
    struct gpfs_file_handle *handle;
    struct gpfs_acl *acl;
    struct ext_cred_t *cred;
    unsigned int access;       /* permission bits */
    unsigned int *supported;	
};

#ifdef __cplusplus
}
#endif

#endif /* H_GPFS_NFS */
